package com.contentsquare.flutter

import android.content.Context
import android.view.View
import com.contentsquare.android.Contentsquare
import com.contentsquare.android.api.bridge.flutter.FlutterInterface
import com.contentsquare.android.api.bridge.flutter.ExternalViewGraphListener
import com.contentsquare.android.api.model.Transaction
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel.Result
import com.contentsquare.android.error.analysis.ErrorAnalysis


private const val NATIVE_EXCEPTION = "NATIVE_EXCEPTION"
private const val MISSING_ARGUMENT = "MISSING_ARGUMENT"
private const val NO_CONTEXT = "NO_CONTEXT"

class MethodCallHandlerImpl {


    fun send(call: MethodCall, result: Result) {
        val screenName = call.argument<String>("screenName")
        if (screenName == null) {
            result.error(MISSING_ARGUMENT, "Missing string for the screenName parameter", null)
            return
        }

        safeExecute(function = {
            Contentsquare.send(screenName)
            result.success(0)
        }, throwableCallback = { error ->
            result.error(NATIVE_EXCEPTION, "Failed to send", error)
        })
    }

    fun sendTransaction(call: MethodCall, result: Result) {
        val price = call.argument<Double>("price")
        val currency = call.argument<String>("currency")
        val transactionId = call.argument<String>("transactionId")

        if (price == null) {
            result.error(MISSING_ARGUMENT, "Missing double for the price parameter", null)
            return
        }
        if (currency == null) {
            result.error(MISSING_ARGUMENT, "Missing string for the currency parameter", null)
            return
        }

        safeExecute(function = {
            if (transactionId == null) {
                Contentsquare.send(Transaction.builder(price.toFloat(), currency).build())
            } else {
                Contentsquare.send(Transaction.builder(price.toFloat(), currency).id(transactionId).build())
            }
            result.success(0)
        }, throwableCallback = { error ->
            result.error(NATIVE_EXCEPTION, "Failed to send transaction", error)
        })
    }

    fun sendDynamicVar(call: MethodCall, result: Result) {
        val key = call.argument<String>("key")
        val stringValue = call.argument<String>("stringValue")
        val intValue = call.argument<Int>("intValue")

        if (key == null) {
            result.error(MISSING_ARGUMENT, "Missing string for the key parameter", null)
            return
        }
        if (stringValue == null && intValue == null) {
            result.error(MISSING_ARGUMENT, "Either stringValue or intValue parameter should be provided", null)
            return
        }

        safeExecute(function = {
            if (stringValue != null) {
                Contentsquare.send(key, stringValue)
            } else if (intValue != null) {
                Contentsquare.send(key, intValue.toLong())
            }
            result.success(0)
        }, throwableCallback = { error ->
            result.error(NATIVE_EXCEPTION, "Failed to send dynamic var", error)
        })
    }

    fun optIn(result: Result, context: Context?) {
        if (context == null) {
            result.error(NO_CONTEXT, "Cannot call optIn without context, please wait for FlutterEngine to start. This can happen if you call the optIn method from the global scope or from outside a build method.", null)
            return
        }

        safeExecute(function = {
            Contentsquare.optIn(context)
            result.success(0)
        }, throwableCallback = { error ->
            result.error(NATIVE_EXCEPTION, "Failed to optIn", error)
        })
    }


    fun optOut(result: Result, context: Context?) {
        if (context == null) {
            result.error(NO_CONTEXT, "Cannot call optOut without context, please wait for FlutterEngine to start. This can happen if you call the optOut method from the global scope or from outside a build method.", null)
            return
        }

        safeExecute(function = {
            Contentsquare.optOut(context)
            result.success(0)
        }, throwableCallback = { error ->
            result.error(NATIVE_EXCEPTION, "Failed to optOut", error)
        })
    }

    fun forgetMe(result: Result) {
        safeExecute(function = {
            Contentsquare.forgetMe()
            result.success(0)
        }, throwableCallback = { error ->
            result.error(NATIVE_EXCEPTION, "Failed to execute forgetMe", error)
        })
    }

    fun getUserId(result: Result) {
        safeExecute(function = {
            val userId = Contentsquare.getUserId()
            result.success(userId)
        }, throwableCallback = { error ->
            result.error(NATIVE_EXCEPTION, "Failed to get user id", error)
        })
    }

    fun stopTracking(result: Result) {
        safeExecute(function = {
            Contentsquare.stopTracking()
            result.success(0)
        }, throwableCallback = { error ->
            result.error(NATIVE_EXCEPTION, "Failed to stop tracking", error)
        })
    }

    fun resumeTracking(result: Result) {
        safeExecute(function = {
            Contentsquare.resumeTracking()
            result.success(0)
        }, throwableCallback = { error ->
            result.error(NATIVE_EXCEPTION, "Failed to resume tracking", error)
        })
    }

    fun currentSessionReplayUrl(result: Result) {
        safeExecute(function = {
            val url = Contentsquare.currentSessionReplayLink()
            result.success(url)
        }, throwableCallback = { error ->
            result.error(NATIVE_EXCEPTION, "Exception occurred when getting URL from native", error)
        })
    }

    fun setURLMaskingPatterns(call: MethodCall, result: Result) {
        val patterns = call.argument<List<String>>("patterns")
        if (patterns == null) {
            result.error("MISSING_ARGUMENT", "Missing list of strings for the patterns parameter", null)
            return
        }

        safeExecute(function = {
            ErrorAnalysis.setUrlMaskingPatterns(patterns)
            result.success(0)
        }, throwableCallback = { error ->
            result.error("NATIVE_EXCEPTION", "Failed to set URL masking patterns", error)
        })
    }

    fun sendTouchEvent(call: MethodCall, result: Result) {
        val body = call.argument<String>("body")

        if (body == null) {
            result.error(MISSING_ARGUMENT, "Missing string for the body parameter", null)
            return
        }

        safeExecute(function = {
            CSFlutter.sendEvent(body)
            result.success(0)
        }, throwableCallback = { error ->
            result.error(NATIVE_EXCEPTION, "Failed to send touch event", error)
        })

    }

    fun collectFlutterVersion(call: MethodCall, result: Result) {
        val version = call.argument<String>("version")

        if (version == null) {
            result.error(MISSING_ARGUMENT, "Missing string for the version parameter", null)
            return
        }

        safeExecute(function = {
            CSFlutter.collectFlutterVersion(version)
            result.success(0)
        }, throwableCallback = { error ->
            result.error(NATIVE_EXCEPTION, "Failed to collect flutter version", error)
        })
    }

    fun collectCustomTelemetryData(call: MethodCall, result: Result) {
        val value = call.argument<String>("value")
        val name = call.argument<String>("name")

        if (name == null) {
            result.error(MISSING_ARGUMENT, "Missing string for the name parameter", null)
            return
        }
        if (value == null) {
            result.error(MISSING_ARGUMENT, "Missing string for the value parameter", null)
            return
        }

        safeExecute(function = {
            CSFlutter.collectCustomTelemetryData(name, value)
            result.success(0)
        }, throwableCallback = { error ->
            result.error(NATIVE_EXCEPTION, "Failed to collect data", error)
        })
    }

    object CSFlutter {

        const val FLUTTER_SEPARATOR = "|flutter|"

        fun registerExternalView(view: View, listener: ExternalViewGraphListener) {
            safeExecute(function = {
                val flutterInterface = FlutterInterface::class.java
                flutterInterface.getDeclaredMethod(
                        "registerExternalView", View::class.java, ExternalViewGraphListener::class.java
                ).run {
                    isAccessible = true
                    invoke(null, view, listener)
                }
            }, throwableCallback = { error ->
                println(error)
            })
        }

        fun unRegisterExternalView(view: View) {
            safeExecute(function = {
                val flutterInterface = FlutterInterface::class.java
                flutterInterface.getDeclaredMethod(
                        "unRegisterExternalView", View::class.java
                ).run {
                    isAccessible = true
                    invoke(null, view)
                }
            }, throwableCallback = { error ->
                println(error)
            })
        }

        fun sendEvent(json: String) {
            val flutterInterface = FlutterInterface::class.java
            flutterInterface.getDeclaredMethod(
                    "sendEvent",
                    String::class.java
            ).run {
                isAccessible = true
                invoke(null, json)
            }
        }

        fun collectFlutterVersion(version: String) {
            val telemetryInterfaceClass = getTelemetryInterfaceClass()
            val instance = telemetryInterfaceClass.newInstance()

            telemetryInterfaceClass.getDeclaredMethod("telemetrySetXPFVersion", Object::class.java).run {
                isAccessible = true
                invoke(instance, version)
            }

            telemetryInterfaceClass.getDeclaredMethod("telemetrySetXPFType", Object::class.java).run {
                isAccessible = true
                invoke(instance, "flutter")
            }
        }

        fun collectCustomTelemetryData(name: String, value: String) {
            val telemetryInterfaceClass = getTelemetryInterfaceClass()
            val instance = telemetryInterfaceClass.newInstance()

            telemetryInterfaceClass.getDeclaredMethod("telemetryCollect", Object::class.java).run {
                isAccessible = true
                invoke(instance, value)
            }
        }

        private fun getTelemetryInterfaceClass(): Class<*> {
            val className = "com.contentsquare.android.api.bridge.telemetry.TelemetryInterface"
            return Class.forName(className)
        }
    }
}
