import Flutter
import UIKit
import ContentsquareModule
import ContentsquareErrorAnalysisModule

// This is the protocol that the listeners passed by external bridges have to implement.
@objc
protocol ExternalBridgeInterface {
    @objc
    func takeSnapshot(parameters: [String: Any])
    
    @objc
    func enableSessionReplay(parameters: [String: Any])

    @objc
    func enableAPIErrors(parameters: [String: Any])
}

private typealias TakeSnapshotHandler = (_ rootPath: String, _ completion: @escaping (String?) -> Void) -> Void
private typealias EnableSessionReplayHandler = (_ enable: Bool) -> Void
private typealias EnableAPIErrorHandler = (_ enable: Bool) -> Void

private class ExternalBridgeImpl: ExternalBridgeInterface {
    private var _takeSnapshotHandler: TakeSnapshotHandler
    private var _enableSessionReplayHandler: EnableSessionReplayHandler
    private var _enableAPIErrorHandler: EnableAPIErrorHandler
    
    init(takeSnapshotHandler: @escaping TakeSnapshotHandler, enableSessionReplayHandler: @escaping EnableSessionReplayHandler, enableAPIErrorHandler: @escaping EnableAPIErrorHandler) {
        _takeSnapshotHandler = takeSnapshotHandler
        _enableSessionReplayHandler = enableSessionReplayHandler
        _enableAPIErrorHandler = enableAPIErrorHandler
    }
    
    @objc
    func takeSnapshot(parameters: [String: Any]) {
        if let rootPath = parameters[TakeSnapshotParam.rootPath.rawValue] as? String,
           let completion = parameters[TakeSnapshotParam.completion.rawValue] as? (String?) -> Void {
            _takeSnapshotHandler(rootPath, completion)
        }
    }
    
    @objc
    func enableSessionReplay(parameters: [String: Any]) {
        if let enable = parameters[EnableSessionReplayParam.enable.rawValue] as? Bool {
            _enableSessionReplayHandler(enable)
        }
    }

    func enableAPIErrors(parameters: [String : Any]) {
        if let enable = parameters[EnableAPIErrorParam.enable.rawValue] as? Bool {
            _enableAPIErrorHandler(enable)
        }
    }
}

// Expected elements takeSnapshot's parameters dictionary:
enum TakeSnapshotParam: String {
    
    // path of the rootView of the bridge on the native side. To be used to create TargetViewPaths when generating the snapshot
    case rootPath // as String - mandatory
    
    // completion block to be used in order to return the json of the generated snapshot asynchronoulsy
    case completion // as (json:String?) -> Void - mandatory
}

enum EnableSessionReplayParam: String {
    
    // flag, that represents the new status of the session recording.
    case enable // as bool - mandatory
}

enum EnableAPIErrorParam: String {

    // flag, that represents the new status of the session recording.
    case enable // as bool - mandatory
}

enum FlutterChannelMethod: String {
    case takeSnapshot = "requestFlutterSnapshot"
    case enableSR = "requestEnableSessionReplay"
    case enableAPIError = "requestEnableAPIError"
}

func registerExternalBridge(channel: FlutterMethodChannel) {
    let flutterTag = "|flutter|"
    
    func snapshotHandler(rootPath: String, completion: @escaping (String?) -> Void) {
        let arguments: [String: Any] = [TakeSnapshotParam.rootPath.rawValue: rootPath + flutterTag]
        channel.invokeMethod(FlutterChannelMethod.takeSnapshot.rawValue, arguments: arguments) { result in
            let rootViewBitmap = rootViewBitmap() ?? ""
            var json = result as? String
            json = json?.replacingOccurrences(of: "com_contentsquare_rootViewBitmapPlaceholder", with: rootViewBitmap)
            completion(json)
        }
    }
    
    func enableSRHandler(enable: Bool) {
        let arguments: [String: Any] = [EnableSessionReplayParam.enable.rawValue: enable]
        channel.invokeMethod(FlutterChannelMethod.enableSR.rawValue, arguments: arguments)
    }

    func enableAPIErrorHandler(enable: Bool) {
        let arguments: [String: Any] = [EnableAPIErrorParam.enable.rawValue: enable]
        channel.invokeMethod(FlutterChannelMethod.enableAPIError.rawValue, arguments: arguments)
     }
    
    let interface = ExternalBridgeImpl(
        takeSnapshotHandler: snapshotHandler,
        enableSessionReplayHandler: enableSRHandler,
        enableAPIErrorHandler: enableAPIErrorHandler
    )
    
    Contentsquare.perform(
        NSSelectorFromString("_registerExternalBridgeWithParameters:"),
        with: [
            "interface": interface,
            "type": 1
        ]
    )
}

private func rootViewBitmap() -> String? {
    let flutterViews = UIApplication.shared.windows.flatMap {
        $0.allSubviews(where: { $0.className == "FlutterView" })
    }
    guard let flutterView = flutterViews.first else {
        return nil
    }
    return UIGraphicsImageRenderer(bounds: flutterView.bounds).image { _ in
        flutterView.drawHierarchy(in: flutterView.bounds, afterScreenUpdates: true)
    }.pngData()?.base64EncodedString()
}

public class SwiftContentsquarePlugin: NSObject, FlutterPlugin {
    private var methodCallHandlerImpl = MethodCallHandlerImpl()
    static private var _channel: FlutterMethodChannel!
    
    public static func register(with registrar: FlutterPluginRegistrar) {
        if (SwiftContentsquarePlugin._channel != nil) {
            NSLog(
            """
            CSLIB ⚠️ WARNING: SwiftContentsquarePlugin was registered twice. Please refer to the Contentsquare documentation's troubleshooting section for more information about what to do: https://docs.contentsquare.com/flutter/#troubleshooting
            """
            );
            return;
        }
        let channel = FlutterMethodChannel(name: "contentsquare", binaryMessenger: registrar.messenger())
        SwiftContentsquarePlugin._channel = channel
        
        let instance = SwiftContentsquarePlugin()
        registrar.addMethodCallDelegate(instance, channel: channel)
        registerExternalBridge(channel: channel)
    }
    
    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        switch call.method {
        case "send":
            methodCallHandlerImpl.send(call, result)
            break
        case "sendTransaction":
            methodCallHandlerImpl.sendTransaction(call, result)
            break
        case "sendDynamicVar":
            methodCallHandlerImpl.sendDynamicVar(call, result)
            break
        case "optIn":
            methodCallHandlerImpl.optIn(call, result)
            break
        case "optOut":
            methodCallHandlerImpl.optOut(call, result)
            break
        case "forgetMe":
            methodCallHandlerImpl.forgetMe(call, result)
            break
        case "getUserId":
            methodCallHandlerImpl.getUserId(call, result)
            break
        case "stopTracking":
            methodCallHandlerImpl.stopTracking(call, result)
            break
        case "resumeTracking":
            methodCallHandlerImpl.resumeTracking(call, result)
            break
        case "sendTouchEvent":
            methodCallHandlerImpl.sendTouchEvent(call, result)
            break
        case "sendSessionReplayEvent":
            methodCallHandlerImpl.sendSessionReplayEvent(call, result)
            break
        case "sendApiError":
            methodCallHandlerImpl.sendApiError(call, result)
            break
        case "collectCustomTelemetryData":
            methodCallHandlerImpl.collectCustomTelemetryData(call, result)
            break
        case "collectFlutterVersion":
            methodCallHandlerImpl.collectFlutterVersion(call, result)
            break
        case "currentSessionReplayUrl":
            methodCallHandlerImpl.currentSessionReplayUrl(result)
            break
        case "setURLMaskingPatterns":
            methodCallHandlerImpl.setURLMaskingPatterns(call, result)
            break
        default:
            result(FlutterMethodNotImplemented)
        }
    }
}


fileprivate let MISSING_ARGUMENT = "MISSING_ARGUMENT"
fileprivate let NATIVE_EXCEPTION = "NATIVE_EXCEPTION"
fileprivate let DYNAMIC_VAR_ERROR = "DYNAMIC_VAR_ERROR"
fileprivate let INVALID_URL = "INVALID_URL"
fileprivate let UNEXPECTED_ERROR = "UNEXPECTED_ERROR"

class MethodCallHandlerImpl {
    
    public func send(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        let arguments: [String : Any] = call.arguments as? [String : Any] ?? [:]
        guard let screenName = arguments["screenName"] as? String else {
            result(FlutterError(
                code: MISSING_ARGUMENT,
                message: "Missing string for the screenName parameter",
                details: nil)
            )
            return
        }
        
        var nativeCustomVars: [CustomVar] = []
        if let customVars = arguments["customVars"] as? [[String: Any]] {
            nativeCustomVars = customVars.compactMap { customVar -> CustomVar? in
                guard let index = customVar["index"] as? NSNumber,
                      let name = customVar["name"] as? String,
                      let value = customVar["value"] as? String else {
                    return nil
                }
                return CustomVar(index: UInt32(index.intValue), name: name, value: value)
            }
        }

        safeExecute(function: {
            () in
            Contentsquare.send(screenViewWithName: screenName, cvars: nativeCustomVars)
            result(nil)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred while sending screenName",
                details: "\(error)")
            )
        });
    }
    
    public func sendTransaction(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        let arguments: [String : Any] = call.arguments as? [String : Any] ?? [:]
        
        let transactionId = arguments["transactionId"] as? String ?? nil
        let priceNumber = arguments["price"] as? NSNumber ?? nil
        let price = priceNumber?.floatValue
        let currency = arguments["currency"] as? String ?? nil
        
        if (price == nil) {
            result(FlutterError(
                code: MISSING_ARGUMENT,
                message: "Missing double for the price parameter",
                details: nil)
            )
            return
        }
        if (currency == nil) {
            result(FlutterError(
                code: MISSING_ARGUMENT,
                message: "Missing string for the currency parameter",
                details: nil)
            )
        }

        safeExecute(function: {
            () in
            Contentsquare.send(transaction: CustomerTransaction(id: transactionId, value: price!, currency: currency!))
            result(nil)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred while sending transaction",
                details: "\(error)")
            )
        });
    }
    
    public func sendDynamicVar(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        let arguments: [String : Any] = call.arguments as? [String : Any] ?? [:]
        
        let key = arguments["key"] as? String ?? nil
        let stringValue = arguments["stringValue"] as? String ?? nil
        let intValue = arguments["intValue"] as? NSNumber ?? nil
        
        if (key == nil) {
            result(FlutterError(
                code: MISSING_ARGUMENT,
                message: "Missing string for the key parameter",
                details: nil)
            )
            return
        }

        if (stringValue == nil && intValue == nil) {
            result(FlutterError(
                code: MISSING_ARGUMENT,
                message: "Either stringValue or intValue parameter should be provided",
                details: nil)
            )
            return
        }

        safeExecute(function: {
            () in
            if (stringValue != nil) {
                Contentsquare.send(dynamicVar: DynamicVar(key: key!, value: stringValue!))
            } else if (intValue != nil) {
                Contentsquare.send(dynamicVar: DynamicVar(key: key!, value: intValue!.uint32Value))
            }
            result(nil)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: DYNAMIC_VAR_ERROR,
                message: "Error while sending the dynamic variable. Please make sure that you have defined the variable in the Contentsquare console",
                details: "\(error)")
            )
        });
    }
    
    public func optIn(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        safeExecute(function: {
            () in
            Contentsquare.optIn()
            result(nil)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred during optIn",
                details: "\(error)")
            )
        });
    }
    
    public func optOut(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        safeExecute(function: {
            () in
            Contentsquare.optOut()
            result(nil)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred during optOut",
                details: "\(error)")
            )
        });
    }
    
    public func forgetMe(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        safeExecute(function: {
            () in
            Contentsquare.forgetMe()
            result(nil)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred during forgetMe",
                details: "\(error)")
            )
        });
    }
    
    public func getUserId(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        safeExecute(function: {
            () in
            result(Contentsquare.userID)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred while getting user ID",
                details: "\(error)")
            )
        });
    }
    
    public func stopTracking(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        safeExecute(function: {
            () in
            Contentsquare.stopTracking()
            result(nil)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred during stopTracking",
                details: "\(error)")
            )
        });
    }
    
    public func resumeTracking(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        safeExecute(function: {
            () in
            Contentsquare.resumeTracking()
            result(nil)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred during resumeTracking",
                details: "\(error)")
            )
        });
    }

    public func currentSessionReplayUrl(_ result: @escaping FlutterResult) {

        safeExecute(function: {
            () in
            let url = Contentsquare.currentSessionReplayLink
            result(url?.absoluteString)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred when getting URL from native",
                details: "\(error)")
            )
        });
    }
    
    public func sendTouchEvent(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        let arguments: [String : Any] = call.arguments as? [String : Any] ?? [:]
        
        let body = arguments["body"] as? String
        
        if (body == nil) {
            result(FlutterError(
                code: MISSING_ARGUMENT,
                message: "Missing string for the body parameter",
                details: nil)
            )
            return
        }

        safeExecute(function: {
            () in
            Contentsquare.perform(NSSelectorFromString("_sendEventWithJson:"), with: body)
            result(nil)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred during sendTouchEvent",
                details: "\(error)")
            )
        });
    }
    
    public func sendSessionReplayEvent(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        let arguments: [String : Any] = call.arguments as? [String : Any] ?? [:]
        
        var body = arguments["body"] as? [String: Any]
        
        if (body == nil) {
            result(FlutterError(
                code: MISSING_ARGUMENT,
                message: "Missing Map for the body parameter",
                details: nil)
            )
            return
        }
        
        safeExecute(function: {
            () in
            Contentsquare.perform(NSSelectorFromString("_sendSessionReplayEventWithParameters:"), with: body)
            result(nil)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred during sendSessionReplayEvent",
                details: "\(error)")
            )
        });
    }

    public func sendApiError(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        let arguments: [String : Any] = call.arguments as? [String : Any] ?? [:]

        safeExecute(function: {
            () in
            let httpMetrics = try DictionaryToHttpMetricConverter.convertFrom(dict: arguments)
            // To be able to send the metric, we need to stop networkMetrics
            httpMetrics.stop();
            result(nil)
        }, errorCallback: {
            (error) in
            let (code, message): (String, String)
                switch error {
                    case HttpMetricConverterError.missingValue(let key):
                        code = MISSING_ARGUMENT
                        message = "Missing value for key '\(key)'"
                    case HttpMetricConverterError.invalidURL(let urlString):
                        code = INVALID_URL
                        message = "Invalid URL '\(urlString)'"
                    default:
                        code = UNEXPECTED_ERROR
                        message = "Unexpected error: '\(error)'"
            }
            result(FlutterError(code: code, message: message, details: nil))
        })
    }

    public func collectCustomTelemetryData(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        guard let arguments = call.arguments as? [String: Any],
              let name = arguments["name"] as? String else {
            result(FlutterError(
                code: MISSING_ARGUMENT,
                message: "Missing String for the name parameter",
                details: nil)
            )
            return
        }
        
        guard let value = arguments["value"] as? String else {
            result(FlutterError(
                code: MISSING_ARGUMENT,
                message: "Missing String for the value parameter",
                details: nil)
            )
            return
        }
        
        safeExecute(function: {
            () in
            let selector = NSSelectorFromString("_telemetryCollect:withValue:")
            if Contentsquare.self.responds(to: selector) {
                Contentsquare.perform(selector, with: name, with: value)
            }
            result(nil)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred during collectCustomTelemetryData",
                details: "\(error)")
            )
        });
    }
    
    public func collectFlutterVersion(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        guard let arguments = call.arguments as? [String: Any],
              let version = arguments["version"] as? String else {
            result(FlutterError(
                code: MISSING_ARGUMENT,
                message: "Missing String for the version parameter",
                details: nil)
            )
            return
        }

        safeExecute(function: {
            () in
            let selectorCollectXpfVersion = NSSelectorFromString("_telemetrySetXPFVersion:")
            if Contentsquare.self.responds(to: selectorCollectXpfVersion) {
                Contentsquare.perform(selectorCollectXpfVersion, with: version)
            }

            let selectorCollectXpfType = NSSelectorFromString("_telemetrySetXPFType:")
            if Contentsquare.self.responds(to: selectorCollectXpfType) {
                Contentsquare.perform(selectorCollectXpfType, with: "flutter")
}
            result(nil)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred during collectCustomTelemetryData",
                details: "\(error)")
            )
        });

    }

    public func setURLMaskingPatterns(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        guard let arguments = call.arguments as? [String: Any],
              let patterns = arguments["patterns"] as? [String] else {
            result(FlutterError(
                code: MISSING_ARGUMENT,
                message: "Missing String array for the patterns parameter",
                details: nil)
            )
            return
        }

        safeExecute(function: {
            () in
            ErrorAnalysis.setURLMaskingPatterns(patterns)
            result(nil)
        }, errorCallback: {
            (error) in
            result(FlutterError(
                code: NATIVE_EXCEPTION,
                message: "Exception occurred during setURLMaskingPatterns",
                details: "\(error)")
            )
        });
    }
}

extension UIView {
    
    func allSubviews(where predicate: (UIView) -> Bool) -> [UIView] {
        let views: [UIView] = predicate(self) ? [self] : []
        return views + subviews.flatMap { $0.allSubviews(where: predicate) }
    }
    
    var className: String {
        String(describing: type(of: self))
    }
}
